static char *RCSid="$Id: hub.c,v 0.1 1996/01/30 01:34:45 dld Exp $";

/* hub [host [port [local-port]]]

   A simple multiplexer for Empire.  The first connecting client
   determines the country name and password.  Further connections must
   use the same name and password.

   -harmless@empire.net */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <netdb.h>
#include <curses.h>
#include <term.h>
#include <fcntl.h>

/* #define DEBUG */

static char *addrname(struct sockaddr_in a) {
  struct hostent *he;
  he = gethostbyaddr((char *)&a.sin_addr, sizeof(a.sin_addr), AF_INET);
  if (he) return he->h_name; else return (inet_ntoa(a.sin_addr));
}

#define COUNTRY  0x1	/* country matches */
#define PASSWORD 0x2	/* password matches */
#define PLAYING  0x4	/* play has been issued */
#define EXITING  0x8	/* client has issued a bye command */

struct client {
  int fd;
  int status;
  int end;		/* end of client input buffer */
  char country[20];
  char password[20];
  char user[20];
  char buffer[1024];
};


#define MAXCLIENTS 100
struct client client[MAXCLIENTS];
static int clients=0;

/* number of lines sent and prompts received back */
/* context is up for grabs when sent==received and last received was a '6'
 * prompt */
static int commands_sent=0,prompts_received=0;

static int focus=-1;	/* the client with the input focus */




/* some defaults in case nothing is specified.  Change as you wish. */
#define DEFPORT "7778"
#define DEFHOST "empire.net"
#define LOCALPORT "1627"


/************************************************************************
 *   The filedescriptor of the empire server socket.  Had to make this
 * global in order to pass to be accessible to event_hook().
 *   I always feel guilty when using globals...
 ************************************************************************/
static int server=-1,local=-1;
static char *host,*port;



/************************************************************************
 * the country and password used to log into the server.
 ************************************************************************/
static char country[20],password[20];
static int btus=0,timeused=0;



/************************************************************
 *    Open a socket to host/port and return filedescriptor.
 ************************************************************/
int empconnect(char *host,int port,char *user) {
  struct	hostent *hp;
  struct	sockaddr_in sin;
  char buffer[1024];
  int s;
  int i;

  printf("connecting to %s/%d as %s\n",host,port,user);

  if (isdigit(*host)) {
    sin.sin_addr.s_addr = inet_addr(host);
  } else {
    hp = gethostbyname(host);
    if (hp == NULL) {
      printf("%s: No such host\n", host);
      return -1;
    }
    memcpy( &sin.sin_addr, hp->h_addr, sizeof(sin.sin_addr));
  }

  sin.sin_port = htons(port);

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0) {
    perror("socket:");
    return -1;
  }

  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&i,sizeof(i))) {
    perror("unable to reuse socket");
  }

  sin.sin_family = AF_INET;
  if (connect(s, &sin, sizeof(sin)) < 0) {
    perror("connect");
    return -1;
  }

  fcntl(s,F_SETFL,FNONBLOCK);


  /*  sprintf(buffer,"user %s\nclient EmpireHub\n",user);
  write(s,buffer,strlen(buffer)); */

  return s;
}

/************************************************************
 *	Handle input lines depending on protocol number
 ************************************************************/
void process_line(char *buffer,int len) {
  int i,j;

  if (len<3) return;

  switch (buffer[0]) {

  case '3':
    fwrite(buffer+2,1,len-2,stderr);
    server=-1;
    break;

  case 'a':
  case 'b':
    fwrite(buffer+2,1,len-2,stderr);
    if (focus>=0) {
      write(client[focus].fd,buffer,len);
    }
    break;

  case '4':
  case '5':
    prompts_received++;
    if (focus>=0) {
      write(client[focus].fd,buffer,len);
    } else {
      fprintf(stderr,"focus error 3(");
      fwrite(buffer,1,len,stderr);
    }
    break;


  case '6':
    prompts_received++;
    sscanf(buffer,"6 %d %d",&btus,&timeused);

    if (focus>=0) {
      write(client[focus].fd,buffer,len);
#ifdef DEBUG
      write(fileno(stderr),buffer,len);
#endif
    } else {
      fprintf(stderr,"focus error 1(");
      fwrite(buffer,1,len,stderr);
    }
    if (commands_sent==prompts_received) {
      focus = -1;
    }
    break;

    /* forward to focus client */
  case '1':
  case '8':
  case '9':
  case 'c':
    if (focus>=0) {
      write(client[focus].fd,buffer,len);
    } else {
      fwrite(buffer,1,len,stdout);
    }
    break;

    
  case '0':
    if (focus>=0) {
      prompts_received++;
      write(client[focus].fd,buffer,len);
    } else {
      fprintf(stderr,"What is this doing here? %s\n",buffer);
    }
    break;

    /* explode to all clients */
  case '2':
  case '7':
  case 'd':
  case 'e':
  case 'f':
    for (i=0;i<clients;i++) {
      write(client[i].fd,buffer,len);
    }
    break;

  default:
    fprintf(stderr,"Unimplemented protocol:\n");
    fwrite(buffer,1,len,stderr);
    break;
  }
}


/* process whole lines, transfer remainder to the top of the buffer
   and return the length of the remainder */
int output(char *buffer,int len) {
  int start,end;

  start=0;
  
  for (;;) {
    for (end=start;end<len;end++) {
      if (buffer[end]=='\n') break;
    }
    if (end<len) {
      process_line(buffer+start,end-start+1);
    } else {
      if (end!=start) {
	memcpy(buffer,buffer+start,end-start);
	buffer[end-start]=0;
	return end-start;
      } else {
	return 0;
      }
    }
    start=end+1;
  }
}





int parse_input_line(char *buf,int len,struct client *client) {
  int i;
  char buffer[256];


#ifdef DEBUG
  printf("parse_input_line(");
  fwrite(buf,len,1,stdout);
#endif

  /* local command */
  if (*buf == '/') {
    fwrite(buf,1,len,stdout);
    return 0;
    
  } else if (!strncmp(buf,"bye",3) || !strncmp(buf,"quit",4)) {
    client->status |= EXITING;
    sprintf(buffer,"1 No worries\n3 so long...\n");
    write(client->fd,buffer,strlen(buffer));
    return 0;

  } else if (client->status & PLAYING) {

#ifdef DEBUG
    printf("sent to server:");
    fwrite(buf,len,1,stdout);
#endif

    write(server,buf,len);

    return 1;

  } else {

    if (len>sizeof(buffer)) {
      write(client->fd,"a line too long\n",16);
      return 0;
    }

    if (sscanf(buf,"coun %s\n",client->country)) {
      if (clients==1) {
	strcpy(country,client->country);
	write(server,buf,len);
	return 1;

      } else {
	if (!strcmp(client->country,country)) {
	  client->status |= COUNTRY;
	  sprintf(buffer,"0 country name %s\n",client->country);
	  write(client->fd,buffer,strlen(buffer));
	} else {
	  sprintf(buffer,"a country %s does not exist\n",client->country);
	  write(client->fd,buffer,strlen(buffer));
	}
	return 0;
      }

    } else if (sscanf(buf,"pass %s\n",client->password)) {
      if (clients==1) {
	strcpy(password,client->password);
	write(server,buf,len);
	return 1;
      } else {
	if (!strcmp(client->password,password)) {
	  client->status |= PASSWORD;
	  sprintf(buffer,"0 password ok\n");
	  write(client->fd,buffer,strlen(buffer));
	} else {
	  sprintf(buffer,"a password bad, logging entry\n");
	  write(client->fd,buffer,strlen(buffer));
	  printf("Bad password %s\n",client->password);
	}
      }
      return 0;

    } else if (sscanf(buf,"user %s\n",client->user)) {
      if (clients==1) {
	write(server,buf,len);
	return 1;

      } else {
	sprintf(buffer,"0 hello %s\n",client->user);
	write(client->fd,buffer,strlen(buffer));
	return 0;
      }

    } else if (!strncmp("play",buf,4)) {
      if (clients==1) {
	client->status |= PLAYING;  /* actually issued too soon */
	write(server,buf,len);
	return 1;

      } else {
	if ((client->status&PASSWORD) && (client->status&COUNTRY)) {
	  client->status |= PLAYING;
	  sprintf(buffer,"2 2\n1 Welcome to the EmpireHub\n6 %d %d\n",
		  btus,timeused);
	  write(client->fd,buffer,strlen(buffer));
	  return 0;

	} else {
	  sprintf(buffer,"a country or password not correct\n");
	  write(client->fd,buffer,strlen(buffer));
	  return 0;

	}
      }
    } else if (!strncmp("client",buf,5)) {

      if (clients==1) {
	write(server,buf,len);
	return 1;

      } else {
	sprintf(buffer,"0 not interested\n");
	write(client->fd,buffer,strlen(buffer));
	return 0;
      }
    }

    fprintf(stderr,"Unable to parse line %s\n",buf);
    return 0;

  }
}


int parse_input(struct client *client) {
  int start=0,sent=0,i;

  for (i=0;i<client->end;i++) {
    if (client->buffer[i]=='\n') {
      sent += parse_input_line(client->buffer+start,i-start+1,client);
      start=i+1;
    }
  }

  memmove(client->buffer,client->buffer+start,client->end-start);
  client->end=0;

  return sent;
}




/* main loop, wait on input and send it to appropriate places */
int input_loop(int local,struct client *client) {
  fd_set in;
  int i,j;
  int len;
  static char buffer[2048];
  char outbuf[2048],buf[2048];
  static int offset=0;	/* unprocessed remnants in buffer */

  /* establish all the fd's to wait on */
  FD_ZERO(&in);
  if (server>=0) FD_SET(server,&in);

  FD_SET(local,&in);

  for (i=0;i<clients;i++) {
    if (focus < 0 || focus==i) {
      FD_SET(client[i].fd,&in);
    }
  }


  i=select( FD_SETSIZE,&in,NULL,NULL,NULL);


  if (i<0) {
    perror("select");
  }

  if (i>0) {

    /* new data returned from server */
    if (server>=0 && FD_ISSET(server,&in)) {

      while ( (len=recv(server,buffer+offset,sizeof(buffer)-offset,0)) > 0) {
#ifdef DEBUG
	write(fileno(stderr),buffer+offset,len);
#endif
	offset=output(buffer,len+offset);
      }

    }


    /* new client trying to connect */
    if (clients<MAXCLIENTS && FD_ISSET(local,&in)) {
      struct sockaddr_in addr;

      bzero(&client[clients],sizeof(*client));
      i=sizeof(addr);
      client[clients].fd = accept(local,(struct sockaddr *)&addr,&i);
      if (client[clients].fd < 0) {
	perror("accept failed");
      } else {
	if (getpeername(client[clients].fd,&addr,&i) < 0) {
	  perror("getpeername error");
	} else {
	  printf("Connection from %s\n",addrname(addr));
	  clients++;

	  if (clients==1 && server<0) { 
	    server=empconnect(host,atoi(port),getenv("USER"));
	    if (server>=0) {
	      printf("Connected to server %s/%s\n",host,port);
	      /* commands_sent += 2; */
	      focus=0;
	    }
	  } else {
	    sprintf(outbuf,"2 Empire server ready\n");
	    write(client[clients-1].fd,outbuf,strlen(outbuf));
	  }
	}
      }
    }

    for (i=0;i<clients;i++) {
      if (FD_ISSET(client[i].fd,&in)) {
	len=read(client[i].fd,client[i].buffer+client[i].end,sizeof(client[i].buffer)-client[i].end-1);

#ifdef DEBUG
	printf("read from %d:",i);
	fwrite(client[i].buffer+client[i].end,len,1,stdout);
#endif

	client[i].end += len;
	if (len>0) {
	  if (focus<=0 || focus==i) {
	    commands_sent += parse_input(client+i);
	    if (commands_sent>prompts_received) {
	      focus=i;
	    } else {
	      focus=-1;
	    }
	  }
	} else {
	  printf("Client %d dropped connection\n",i);
	  if (i==focus) focus=-1;
	  close(client[i].fd);
	  printf("Copying client %d to %d\n",clients-1,i);
	  client[i]=client[clients-1];
	  clients--;
	  i--;
	}
      }
    }
  }

  /* drop connection to any polite clients that issued bye commands */
  for (i=0;i<clients;i++) {
    if (client[i].status & EXITING) {
      printf("Client %d exiting politely\n",i);
      if (i==focus) focus=-1;
      close(client[i].fd);
      client[i]=client[clients-1];
      clients--;
      i--;
    }
  }

  return clients;
}



/***********************************************************************
 *    Set up readline() and socket s to server.  Push each new line
 *  down the socket.
 ***********************************************************************/
int main(int argc, char **argv)
{
  char *localport;
  int i;
  int persistent=1;		/* server sticks around even when no clients */
  int local=-1;

  if (argc>1) {
    host=argv[1];
  } else {
    host=DEFHOST;
  }

  if (argc>2) {
    port=argv[2];
  } else {
    port=DEFPORT;
  }

  if (argc>3) {
    localport=argv[3];
  } else {
    localport=LOCALPORT;
  }

  local=socket(PF_INET,SOCK_STREAM,getprotobyname("tcp")->p_proto);

  if (local<0) {
    perror("unable to create socket");
    exit(-1);
  }

  if (setsockopt(local,SOL_SOCKET,SO_REUSEADDR,(char *)&i,sizeof(i))) {
    perror("unable to reuse socket");
  }

  {
    struct sockaddr_in addr;

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(atoi(localport));
    if (bind(local, (struct sockaddr *)&addr, sizeof(addr))) {
      perror("bind error");
      exit(errno);
    }
  }

  if (listen(local,5)) {
    perror("listen failed");
  }


  do {
    while (clients=input_loop(local,client)) {}

    printf("Closing server connection\n");
    close(server);
    server=-1;

  } while (persistent);

  close(local);
  return 0;
}



